<?php

namespace Drupal\sdc\Twig;

use Drupal\Component\Utility\Html;
use Drupal\Core\Render\Element;
use Drupal\Core\Template\Attribute;
use Drupal\sdc\ComponentPluginManager;
use Drupal\sdc\Exception\ComponentNotFoundException;
use Drupal\sdc\Exception\InvalidComponentException;
use Drupal\sdc\Plugin\Component;
use JsonSchema\Validator;
use Twig\Extension\AbstractExtension;
use Twig\TwigFunction;

/**
 * The twig extension so Drupal can recognize the new code.
 */
class TwigExtension extends AbstractExtension {

  /**
   * Creates TwigExtension.
   *
   * @param \Drupal\sdc\ComponentPluginManager $pluginManager
   *   The plugin manager.
   */
  public function __construct(private readonly ComponentPluginManager $pluginManager) {
  }

  /**
   * {@inheritdoc}
   */
  public function getNodeVisitors(): array {
    return [new ComponentNodeVisitor()];
  }

  /**
   * {@inheritdoc}
   */
  public function getFunctions(): array {
    return [
      new TwigFunction(
        'sdc_additional_context',
        [$this, 'addAdditionalContext'],
        ['needs_context' => TRUE]
      ),
      new TwigFunction(
        'sdc_validate_props',
        [$this, 'validateProps'],
        ['needs_context' => TRUE]
      ),
    ];
  }

  /**
   * Appends additional context to the template based on the template id.
   *
   * @param array &$context
   *   The context.
   * @param string $component_id
   *   The component ID.
   *
   * @throws \Drupal\sdc\Exception\ComponentNotFoundException
   */
  public function addAdditionalContext(array &$context, string $component_id): void {
    $component = $this->pluginManager->find($component_id);
    $context = array_merge(
      $context,
      $this->computeAdditionalRenderContext($component)
    );
  }

  /**
   * Calculates additional context for this template.
   *
   * @param \Drupal\sdc\Plugin\Component $component
   *   The component.
   *
   * @return array
   *   The additional context to inject to component templates.
   */
  protected function computeAdditionalRenderContext(Component $component): array {
    $metadata = $component->getMetadata();
    $status = $metadata->getStatus();
    $machine_name = $component->getMachineName();
    $classes = array_map([Html::class, 'cleanCssIdentifier'], [
      'sdc',
      'sdc--' . $machine_name,
      'sdc--' . $status,
    ]);
    $classes = array_map('strtolower', $classes);
    $attributes = [
      'class' => $classes,
      'data-sdc-id' => $component->getPluginId(),
    ];
    return [
      'clAttributes' => new Attribute($attributes),
      'clMeta' => $metadata->normalize(),
    ];
  }

  /**
   * Validates the props in development environments.
   *
   * @param array $context
   *   The context provided to the component.
   * @param string $component_id
   *   The component ID.
   *
   * @throws \Drupal\sdc\Exception\InvalidComponentException
   */
  public function validateProps(array &$context, string $component_id): void {
    assert($this->doValidateProps($context, $component_id));
  }

  /**
   * Performs the actual validation of the schema for the props.
   *
   * @param array $context
   *   The context provided to the component.
   * @param string $component_id
   *   The component ID.
   *
   * @return bool
   *   TRUE if it's valid.
   *
   * @throws \Drupal\sdc\Exception\InvalidComponentException
   */
  public function doValidateProps(array $context, string $component_id): bool {
    // Get the props from the context.
    try {
      $component = $this->pluginManager->find($component_id);
    }
    catch (ComponentNotFoundException $e) {
      throw new InvalidComponentException($e->getMessage(), $e->getCode(), $e);
    }
    $props_schema = $component->getMetadata()->getSchemas()['props'] ?? [];
    $prop_names = array_keys($props_schema['properties'] ?? []);
    $props = Validator::arrayToObjectRecursive(array_intersect_key($context, array_flip($prop_names)));
    $throw_error = static function (array $errors) {
      $message_parts = array_map(
        static fn(array $error): string => sprintf("[%s] %s", $error['property'], $error['message']),
        $errors
      );
      $message = implode("/n", $message_parts);
      throw new InvalidComponentException($message);
    };
    $validator = new Validator();
    $validator->validate($props, Validator::arrayToObjectRecursive($props_schema));
    $validator->getErrors();
    if ($validator->isValid()) {
      return TRUE;
    }
    // Dismiss type errors if the prop received a render array.
    $errors = array_filter(
      $validator->getErrors(),
      function (array $error) use ($context): bool {
        if (($error['constraint'] ?? '') !== 'type') {
          return TRUE;
        }
        return !$this->isRenderArray($context[$error['property']] ?? NULL);
      }
    );
    if (empty($errors)) {
      return TRUE;
    }
    $throw_error($errors);
  }

  /**
   * Checks if a candidate is a render array.
   *
   * @param mixed $candidate
   *   The candidate.
   *
   * @return bool
   *   TRUE if it's a render array. FALSE otherwise.
   */
  protected function isRenderArray($candidate): bool {
    if (!is_array($candidate)) {
      return FALSE;
    }
    $children_keys = Element::children($candidate);
    if (empty($children_keys)) {
      // This is a render array, if there is at least one non-children element.
      return !empty($candidate);
    }
    foreach ($children_keys as $key) {
      if (!$this->isRenderArray($candidate[$key] ?? NULL)) {
        return FALSE;
      }
    }
    return TRUE;
  }

}
